#!/usr/bin/python
#    SSHerminator - Terminator with SSH extensions
#
#    Copyright (C) 2008 David Futcher <bobbo@ubuntu.com>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 2 only.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

try:
    import pygtk
    pygtk.require('2.0')
except ImportError:
    pass
import gtk
import os
import yaml
import datetime
import sys
from gettext import ngettext

class ConnectDialog:
  """ Main connection dialog. Opened when user requests and SSH enabled split """

  def __init__ (self, vertical=False, replace=False):
    """ Sets up the GTK part of the dialog """
    
    self.window = gtk.Window()
    self.window.set_title("Create new SSH connection")
    self.window.set_default_size(300, 300)
    self.box = gtk.VBox()
    self.inputbox = gtk.VBox()
    self.buttonbox = gtk.HBox(homogeneous=True)
    self.inputalign = gtk.Alignment(xalign=0.0)
    self.window.data = SSHConnection(vertical=vertical, replace=replace)
    self.connect_history = SSHConnectionHistory()
    self.known_hosts = OpenSSHKnownHosts()
    self.edit_lock = False
    self.window.cancelled = True # Assume user hasnt clicked connect to stop invalid error messages

    # Set up the submit buttons
    self.submit_button = gtk.Button(stock=gtk.STOCK_CONNECT)
    self.submit_button.connect('clicked', self.apply)
    self.buttonbox.add(self.submit_button)
    
    self.cancel_button = gtk.Button(stock=gtk.STOCK_CANCEL)
    self.cancel_button.connect('clicked', self.cancel)
    self.buttonbox.add(self.cancel_button)
    
    # Now create an entrybox and a label for each one
    self.host_container = gtk.HBox(homogeneous=True)
    self.host_box = gtk.Entry()
    self.host_label = gtk.Label("Hostname: ")
    self.host_label.set_alignment(xalign=0.1, yalign=0.5) 
    self.host_container.add(self.host_label)
    self.host_container.add(self.host_box)
    
    self.user_container = gtk.HBox(homogeneous=True)
    self.user_box = gtk.Entry()
    self.user_box.set_text(self._get_username())
    self.user_label = gtk.Label("Username: ")
    self.user_label.set_alignment(xalign=0.1, yalign=0.5)
    self.user_container.add(self.user_label)
    self.user_container.add(self.user_box)
    
    self.port_container = gtk.HBox(homogeneous=True)
    self.port_box = gtk.SpinButton(gtk.Adjustment(lower=1, upper=sys.maxint, step_incr=1), 1.0, 0)
    self.port_box.set_numeric(True)
    self.port_box.set_value(22)
    self.port_label = gtk.Label("Port: ")
    self.port_label.set_alignment(xalign=0.1, yalign=0.5)
    self.port_container.add(self.port_label)
    self.port_container.add(self.port_box)
    
    self.opts_container = gtk.HBox(homogeneous=True)
    self.opts_box = gtk.Entry()
    self.opts_label = gtk.Label("SSH Options: ")
    self.opts_label.set_alignment(xalign=0.1, yalign=0.5)
    self.opts_container.add(self.opts_label)
    self.opts_container.add(self.opts_box)
    
    # Add all the buttons into the window
    self.inputbox.add(self.host_container)
    self.inputbox.add(self.user_container)
    self.inputbox.add(self.port_container)
    self.inputbox.add(self.opts_container)
    self.inputalign.add(self.inputbox)
    self.box.pack_start(self.inputalign, False, True)
    
    # Now for the complicated Treeview stuff
    self.box.pack_start(gtk.Label(), False, True) # Is this an OK way to make some extra space?
        
    self.treestore = gtk.TreeStore(str, object)
              
    # Connection history data
    self.hist_parent = self.treestore.append(None, ["History", None])
    for host in self.connect_history.connections_list:
      self.treestore.append(self.hist_parent, [str(host), host])
      
    self.known_parent = self.treestore.append(None, ["OpenSSH Known Hosts", None])
    for host in self.known_hosts.hosts:
      # Create an SSH connection object for this so we can use the same code
      # to load an OpenSSH host and an old connection from History
      self.treestore.append(self.known_parent, [str(host).split(' ')[0], SSHConnection(host=host)])
    
    self.scrolledwindow = gtk.ScrolledWindow()
    self.scrolledwindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
    self.treeview = gtk.TreeView(self.treestore)
    self.treeview.add_events(gtk.gdk.BUTTON_PRESS_MASK)
    self.treeview.connect('button_press_event', self.right_row_click)
    self.treeview.connect('cursor-changed', self.row_single_click)
    
    self.tvcolumn = gtk.TreeViewColumn('Other hosts')
    self.treeview.append_column(self.tvcolumn)
    self.cell = gtk.CellRendererText()
    self.tvcolumn.pack_start(self.cell, True)
    self.tvcolumn.add_attribute(self.cell, 'text', 0)
    self.treeview.set_search_column(0)
    self.tvcolumn.set_sort_column_id(0)
    self.treeview.set_reorderable(True)
    self.scrolledwindow.add(self.treeview)
    self.box.pack_start(self.scrolledwindow, True, True)
    
    # Stick it all in the window
    self.box.pack_start(gtk.Label(), False, True)
    self.box.pack_start(self.buttonbox, False, True)
    self.window.add (self.box)
        
  def _get_username(self):
    """ Safely gets the current users username from evironment variables """
    try:
      return os.environ['USERNAME']
    except:
      try:
        return os.environ['LOGNAME']
      except:
        return ""
       
  def row_single_click(self, treeview):
    try:
      treesel = treeview.get_selection()
      (model, i) = treesel.get_selected()
      self.data = self.treestore.get_value(i, 1)
    except:
      return 
  
    if self.treestore.is_ancestor(self.hist_parent, i) or self.treestore.is_ancestor(self.known_parent, i):
      # Is in connection history, or is an OpenSSH known host
      if not self.data.host == None:
        self.host_box.set_text(self.data.host)

      if not self.data.user == None:
        self.user_box.set_text(self.data.user)
        
      if not self.data.port == None:
        self.port_box.set_text(self.data.port)
        
      if not self.data.opts == None:
        self.opts_box.set_text(self.data.opts) 
    else:
      # User clicked on a parent row, do nothing
      return
    
  def right_row_click(self, treeview, event):
    """ Called when a row in the treeview (History/Known Hosts) is clicked. Job is to
        find out what row was clicked and put the correct data into the textboxes """
   
    try:
      treesel = treeview.get_selection()
      (model, i) = treesel.get_selected()
      self.data = self.treestore.get_value(i, 1)
    except:
      return 
      
    if event.button == 3:
      # Right click, if a history object create a context menu
      if self.treestore.is_ancestor(self.hist_parent, i):
        self.cur_obj = i
        menu = self._treeview_context_menu()
        menu.show_all()
        menu.popup(None, None, None, event.button, event.time)
      else:
        return # OpenSSH know host. Currently no operations can be applied to the object
      
  def _treeview_context_menu(self):
    """ Generates the context menu for editing a connection history entry """
    menu = gtk.Menu()
    
    edit = gtk.ImageMenuItem(gtk.STOCK_EDIT)
    edit.connect('activate', self._edit_hist_obj)
    menu.append(edit)
    
    edit = gtk.ImageMenuItem(gtk.STOCK_DELETE)
    edit.connect('activate', self._delete_hist_obj)
    menu.append(edit)
    
    return menu
            
  def _edit_hist_obj(self, event):
    """ Handler for the edit button in the history context menu. Shows the edit
        dialog, but doesnt handle anything after the dialog is shown """
    
    # We cant have more than one edit dialog open at once (otherwise we get some
    # big race conditions. Time for some basic locking.
    if self.edit_lock:
      return 
    else:
      self.edit_lock = True
      
    dlg = EditConnectionDialog(self.data)
    dlg.window.connect('destroy', self._hist_edit_finished)
    dlg.window.show_all()
    
  def _hist_edit_finished(self, event):
    """ Handles the connection edit process after the new data has been entered
        in the edit dialog """
    try:
      c = event.connection
      d = event.orig_connection
    except AttributeError:
      # User has pressed cancel, remove the lock and return. 
      self.edit_lock = False
      return
      
    # Update the backend
    self.connect_history.delete_connection(d.host, d.user, d.port, d.opts)
    self.connect_history.add_connection(c)
    
    # Update the treestore
    self.treestore.insert_after(self.hist_parent, self.cur_obj, [str(c)])
    self.treestore.remove(self.cur_obj)
    
    # Update the actual data entry fields?  
    
    # Update the lock
    self.edit_lock = False      
    
  def _delete_hist_obj(self, event):
    """ Handler for when user clicks on the Delete button in the history context
        menu. Deletes it from the YAML file, then removes it from the treeview """
    self.connect_history.delete_connection(self.data.host, self.data.user, self.data.port, self.data.opts)
    self.connect_history.write()
    self.treestore.remove(self.cur_obj)
            
  def apply(self, event):
    """ Called when the apply button is clicked, sets up the ssh_data dictionary and
        then destroys the window (destoy signal is caught which lets us create the
        ssh terminal """
    self.window.data.host = self.host_box.get_text()
    self.window.data.user = self.user_box.get_text()
    self.window.data.port = self.port_box.get_text()
    self.window.data.opts = self.opts_box.get_text()
           
    # Add these connection details to the connection history
    self.connect_history.add_connection(self.window.data)
    self.window.cancelled = False
    self.window.destroy()
    
  def cancel(self, event):
    """ Called when user clicks cancel """
    self.window.destroy()
    
class EditConnectionDialog:
  """ Dialog for editing an SSHConnection object """
   
  def __init__(self, connection):
    """ Takes an SSHConnection object and sets up with window """
    self.connection = connection
    self.window = gtk.Window()
    self.window.set_title("Edit connection")
    
    self.box = gtk.VBox()
    self.inputbox = gtk.VBox()
    self.buttonbox = gtk.HBox(homogeneous=True)
    self.inputalign = gtk.Alignment(xalign=0.0)
      
    # Now create an entrybox and a label for each one
    self.host_container = gtk.HBox(homogeneous=True)
    self.host_box = gtk.Entry()
    self.host_box.set_text(self.connection.host)
    self.host_label = gtk.Label("Hostname: ")
    self.host_label.set_alignment(xalign=0.1, yalign=0.5) 
    self.host_container.add(self.host_label)
    self.host_container.add(self.host_box)
    
    self.user_container = gtk.HBox(homogeneous=True)
    self.user_box = gtk.Entry()
    self.user_box.set_text(self.connection.user)
    self.user_label = gtk.Label("Username: ")
    self.user_label.set_alignment(xalign=0.1, yalign=0.5)
    self.user_container.add(self.user_label)
    self.user_container.add(self.user_box)
    
    self.port_container = gtk.HBox(homogeneous=True)
    self.port_box = gtk.SpinButton(gtk.Adjustment(lower=1, upper=sys.maxint, step_incr=1), 1.0, 0)
    self.port_box.set_numeric(True)
    self.port_box.set_value(int(self.connection.port))
    self.port_label = gtk.Label("Port: ")
    self.port_label.set_alignment(xalign=0.1, yalign=0.5)
    self.port_container.add(self.port_label)
    self.port_container.add(self.port_box)
    
    self.opts_container = gtk.HBox(homogeneous=True)
    self.opts_box = gtk.Entry()
    self.opts_box.set_text(self.connection.opts)
    self.opts_label = gtk.Label("SSH Options: ")
    self.opts_label.set_alignment(xalign=0.1, yalign=0.5)
    self.opts_container.add(self.opts_label)
    self.opts_container.add(self.opts_box)
    
    # Add all the buttons into the window
    self.inputbox.add(self.host_container)
    self.inputbox.add(self.user_container)
    self.inputbox.add(self.port_container)
    self.inputbox.add(self.opts_container)
    self.inputalign.add(self.inputbox)
    self.box.add(self.inputalign)
    
    # Set up the submit buttons
    self.submit_button = gtk.Button(stock=gtk.STOCK_APPLY)
    self.submit_button.connect('clicked', self.apply)
    self.buttonbox.add(self.submit_button)
    
    self.cancel_button = gtk.Button(stock=gtk.STOCK_CANCEL)
    self.cancel_button.connect('clicked', self.cancel)
    self.buttonbox.add(self.cancel_button)
    self.box.add(self.buttonbox)
    
    self.window.add(self.box)

  def apply(self, event):
    # We have to put this in the window class so it will show up in the event. HBIW
    self.window.connection = SSHConnection(self.host_box.get_text(), self.user_box.get_text(),
                                           self.port_box.get_text(), self.connection.date,
                                           self.opts_box.get_text())
    self.window.orig_connection = self.connection
    self.window.destroy()
  
  def cancel(self, event):
    self.window.destroy()
      
class SSHConnection:
  """ SSHConnection class represents one old connection in the connections history """
  
  def __init__(self, host = None, user = None, port = 22, date = None, opts = None, vertical = False, 
replace=False):
    """ Get variables setup, easier than having it all in a big dict """
    self.host = host
    self.user = user
    self.port = port
    self.date = date
    self.opts = opts
    self.command = None
    self.vertical = vertical
    self.replace = replace
    self.enabled = False
    
  def __str__(self):
    """ Returns a user friendly string describing the connection """
    return "%s@%s:%s - Last connect %s" % (self.user, self.host, self.port, self.time_since())
    
  def time_since(self):
    """ Calculates the time since the last connection using these settings was made """
    # Parse the date string into a datetime format (why the hell is this not in stdlib?)
    # Here be some mighty dragons
    d = self.date
    d = d.split()
    ymd = d[0].split("-")
    year = ymd[0]; month = ymd[1]; day = ymd[2]
    
    time = d[1].split(":")
    hour = time[0]; minute = time[1]; second = time[2].split(".")[0]; msec = time[2].split(".")[1]
    
    dt = datetime.datetime(int(year), int(month), int(day), int(hour), int(minute), int(second), int(msec))
    
    return timesince(dt, now=datetime.datetime.now())
    
  def bump_date(self):
    self.date = str(datetime.datetime.now())
    
class SSHConnectionHistory:
  """ Handles the backend functions for the SSH connections history """
  
  def __init__(self):
    """ Sets up the class. Just loads up the data """
    self.connections_list = self._load_data()
    
  def _load_data(self):
    """ Loads data from the YAML file and puts it into a list for use in the connections
        dialog """
    self.datadir = "%s/.ssherminator" % os.environ['HOME']
    self.datafile = "%s/connection_history.yaml" % self.datadir
    
    if not os.path.exists(self.datafile):
      if not os.path.isdir(self.datadir):
        os.mkdir(self.datadir)
      
      os.system("touch %s" % self.datafile) # TODO: Do this in Python
      
    handler = open(self.datafile, 'r')
    conn_list = []
    
    try:
      connections = yaml.load(handler.read())['connections']
    except:
      return []
    
    try:
      for conn in connections:
        tmp = SSHConnection(conn['host'], conn['user'], conn['port'], conn['date'], conn['opts'])
        conn_list.append(tmp)
    except:
      print "Malformed connections history file. Ignoring"
      conn_list = []
      
    return conn_list
    
  def _generate_yaml_dict(self):
    """ Generates dictionaries for all the SSHConnection objects in the connections
        list, so the YAML module can dump them into the file """
    conns = []
    
    for c in self.connections_list:
      conns.append({"host" : c.host, "user" : c.user, "port" : c.port, "date" : c.date, "opts" : c.opts})
      
    return {"connections" : conns}
    
  def write(self):
    """ Writes all of the current connections in the connections list into the YAML
        file """
    data = self._generate_yaml_dict()
    yaml.dump(data, open(self.datafile, 'w'))
    
  def add_connection(self, conn):
    """ Adds a connection to the connections list, then writes it to the YAML file """
    conn.bump_date() # Bump the last connection date to now
    
    # Check for duplicates
    for c in self.connections_list:
      if c.host == conn.host and c.user == conn.user and c.port == conn.port and c.opts == conn.opts:
        self.write()
        return 0
    
    self.connections_list.append(SSHConnection(conn.host, conn.user, conn.port, conn.date, conn.opts))
    self.write() # Immediately dump the data back into the file (but dont reload it, because we have already put it into
                 # the connections list (would just add extra weight to the app)
                 
  def get_connection(self, host = None, user = None, port = 22):
    """ Returns the connection object that matches the host, username and port 
        (needed so we can get the SSH commandline options into the dialog) """
    for c in self.connections_list:
      if c.host == host and c.user == user and c.port == port:
        return c
        
  def delete_connection(self, host = None, user = None, port = 22, opts = None):
    """ Delets a connection from the connections history """
    for i in range(0, len(self.connections_list)):
      try:
        c = self.connections_list[i]
        if c.host == host and c.user == user and c.port == port and c.opts == opts:
          del self.connections_list[i]      
      except:
        pass
    
class OpenSSHKnownHosts:
  """ Class to handle the OpenSSH Known Hosts file """
  """ TODO: Check in /etc/ssh/*options* if known_hosts hashing is turned on, 
            and tell them this in the treeview """

  def __init__(self):
    """ Prepares the list for reading by the treeview """
    self.hosts = []
    
    self._read_hosts()

    if len(self.hosts) == 0:
      self.hosts.append("None")
    
  def _read_hosts(self):
    """ Reads the hosts from the known_hosts file and puts them into a list """
    try:
      f = open("%s/.ssh/known_hosts" % os.environ['HOME'])
    except:
      self.hosts.append("None")
      return

    for line in f:
      if not line or line.startswith('|'):
        continue
      
      self.hosts.append(line.split(",")[0])
      
    f.close()
    self.hosts.sort()

    
# This is lifted from Django:
# http://code.djangoproject.com/browser/django/trunk/django/utils/timesince.py
#Copyright (c) 2005, the Lawrence Journal-World
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without modification,
#are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice, 
#       this list of conditions and the following disclaimer.
#    
#    2. Redistributions in binary form must reproduce the above copyright 
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of Django nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
#THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
#ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


def timesince(d, now=None):
    """
    Takes two datetime objects and returns the time between d and now
    as a nicely formatted string, e.g. "10 minutes".  If d occurs after now,
    then "0 minutes" is returned.

    Units used are years, months, weeks, days, hours, and minutes.
    Seconds and microseconds are ignored.  Up to two adjacent units will be
    displayed.  For example, "2 weeks, 3 days" and "1 year, 3 months" are
    possible outputs, but "2 weeks, 3 hours" and "1 year, 5 days" are not.

    Adapted from http://blog.natbat.co.uk/archive/2003/Jun/14/time_since
    """
    chunks = (
      (60 * 60 * 24 * 365, lambda n: ngettext('year', 'years', n)),
      (60 * 60 * 24 * 30, lambda n: ngettext('month', 'months', n)),
      (60 * 60 * 24 * 7, lambda n : ngettext('week', 'weeks', n)),
      (60 * 60 * 24, lambda n : ngettext('day', 'days', n)),
      (60 * 60, lambda n: ngettext('hour', 'hours', n)),
      (60, lambda n: ngettext('minute', 'minutes', n))
    )

    if now:
        t = now.timetuple()
    else:
        t = time.localtime()
    tz = None
    now = datetime.datetime(t[0], t[1], t[2], t[3], t[4], t[5], tzinfo=tz)

    # ignore microsecond part of 'd' since we removed it from 'now'
    delta = now - (d - datetime.timedelta(0, 0, d.microsecond))
    since = delta.days * 24 * 60 * 60 + delta.seconds
    if since <= 0:
        # d is in the future compared to now, stop processing.
        return _('in the future')
    if since < 60:
        return _('less than 1 minute ago')    
    for i, (seconds, name) in enumerate(chunks):
        count = since // seconds
        if count != 0:
            break
    s = None   
    if i + 1 < len(chunks):
        # Now get the second item
        seconds2, name2 = chunks[i + 1]
        count2 = (since - (seconds * count)) // seconds2
        if count2 != 0:        
            s = _('%(number)d %(type)s, %(number2)d %(type2)s ago') % {'number': count, 'type': name(count), 
                                                                       'number2': count2, 'type2': name2(count2)}        
    if not s:
        s = _('%(number)d %(type)s ago') % {'number': count, 'type': name(count)}        
    return s
    
